/*///////////////////////////////////////////////////////////////////////////////
// \file       base64.cpp
    Implementation file for functions to encode and decode base64 (RFC 4648)
    According to http://tools.ietf.org/html/rfc4648

    Version:    1.0                       <br>
    Copyright:  Orjan Westin 2011         <br>
    Licence:    BSD (see EOF comment)     <br>
    Contact:    orjan@coolcowstudio.co.uk <br>
    Download:   http://coolcowstudio.co.uk/source/cpp/coding.zip
//////////////////////////////////////////////////////////////////////////////*/

#include "base64.h"
#include <cassert>

namespace
{
  const char* Alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";

  typedef __int32 bitbucket;
  typedef unsigned char index;

  // There are two bits left over in an index, so we use them to indicate non-index results
  const index Illegal = 0xc0; 
  const index Padding = 0x80;
  const index CrLf    = 0x40;

  const char CR       = 0x0d;
  const char LF       = 0x0a;

  template<typename T>
  char index_to_char(index i)
  {
    // Only six lower bytes permitted in index
    assert(0 == (i & Illegal));

    switch (i)
    {
    case 62:
      return T::char62();
    case 63:
      return T::char63();
    default:
      return Alphabet[i];
    }
  }

  // Look up given character in index
  // Expects only characters from alphabet, plos template-given 62 and 62, padding, and CRLF
  template<typename T>
  index char_to_index(char c)
  {
    // Start with special cases

    // Padding allowed and present?
    if (T::pad() && (T::charPad() == c))
      return Padding;
    // Line break allowed and present?
    if (T::lineBreaks() && ((CR == c) || (LF == c)))
      return CrLf;

    // Look for proper index
    if (T::char62() == c)
      return 62;
    if (T::char63() == c)
      return 63;
    // Look up in alphabet
    for (index i = 0; i < 62; ++i)
      if (Alphabet[i] == c)
          return i;

    // Illegal character should either be ignored or cause error
    if (T::ignoreIllegal())
      return Illegal;
    throw std::invalid_argument("Illegal character");
  }

}

namespace CoCoSt 
{
  //! Utility function namespace
  namespace Coding
  {
    // Declare the template functions with the types we are prepared to support
    // This instantiates these functions for these types, so they can be linked
    // Downside is every version of the function is compiled separately, so code grows
    // Therefore, comment out or delete the variants you do not need

    template void bytes_to_base64<base64>(const unsigned char* data, size_t length, std::string& str);
    template void bytes_to_base64<base64>(const std::vector<unsigned char>& data, std::string& str);
    template void base64_to_bytes<base64>(const std::string& str, std::vector<unsigned char>& data);

    template void bytes_to_base64<base64MIME>(const unsigned char* data, size_t length, std::string& str);
    template void bytes_to_base64<base64MIME>(const std::vector<unsigned char>& data, std::string& str);
    template void base64_to_bytes<base64MIME>(const std::string& str, std::vector<unsigned char>& data);

    template void bytes_to_base64<base64url>(const unsigned char* data, size_t length, std::string& str);
    template void bytes_to_base64<base64url>(const std::vector<unsigned char>& data, std::string& str);
    template void base64_to_bytes<base64url>(const std::string& str, std::vector<unsigned char>& data);

    template void bytes_to_base64<base64PEM>(const unsigned char* data, size_t length, std::string& str);
    template void bytes_to_base64<base64PEM>(const std::vector<unsigned char>& data, std::string& str);
    template void base64_to_bytes<base64PEM>(const std::string& str, std::vector<unsigned char>& data);

    // Encode
    template<typename T>
    void bytes_to_base64(const unsigned char* data, size_t length, std::string& str)
    {
      str.clear();
      // Try to get max expected size (bytes, padding, line breaks) 
      str.reserve((length * 4 ) / 3 + 3 + (T::lineBreaks() ? 2 * length / T::maxLineLength() : 0) );

      // In case we care about line lengths, keep track of it
      int lineLength = 0;
      for (size_t i = 0; i < length; )
      {
        // Multiple loops, might as well have common iterator
        int a;
        // In this iteration, do we have a full set or are we at the end?
        int byteCount = (3 > (length - i)) ? length - i : 3;

        // Each iteration will generate four characters, so check line length first
        if (T::lineBreaks() && (lineLength + 4 > T::maxLineLength()))
        {
          str += CR;
          str += LF;
          lineLength = 0;
        }
        // Fill bucket of bits
        bitbucket bits = 0;
        for (a = 0; a < byteCount; ++a)
        {
          bits = (bits << 8) | data[i++];
        }
        // Shift up to fill bucket if data is too short
        for (a = byteCount; a < 3; ++a)
        {
          bits = (bits << 8);
        }
        // Extract indices
        // There can be 1-3 bytes in source, which corresponds to 2-4 6-bit indices
        static int shiftlength[] = {18, 12, 6, 0};
        for (a = 0; a <= byteCount; ++a)
        {
          index ind = 0x3f & (bits >> shiftlength[a]);
          str += index_to_char<T>(ind);
          ++lineLength;
        }
        // Padding at the end?
        for (a = byteCount; T::pad() && (a < 3); ++a)
        {
          str += T::charPad();
          // No point in counting up line length, as this is the last iteration
        }
      }
    }

    // Forward call to encode
    template<typename T>
    void bytes_to_base64(const std::vector<unsigned char>& data, std::string& str)
    {
      bytes_to_base64<T>(&data[0], data.size(), str);
    }

    // Decode
    template<typename T>
    void base64_to_bytes(const std::string& str, std::vector<unsigned char>& data)
    {
      data.clear();
      // Not much point in making this exact, just big enough
      data.reserve((str.length() * 3) / 4 + 3);

      // In case we care about line lengths, keep track of it
      int lineLength = 0;

      for (size_t i = 0; i < str.length(); )
      {
        // Make sure we haven't exceeded maximum line length
        if (T::lineBreaks() && (T::maxLineLength() < lineLength))
          throw std::invalid_argument("Line too long");

        // Fill bucket of bits
        bitbucket bits = 0;

        int chars = 0;
        while ((chars < 4) && (i < str.length()))
        {
          index ind = char_to_index<T>(str[i++]);
          ++lineLength;
          // Add legal characters
          if (0 == (Illegal & ind))
          {
            bits = (bits << 6) | ind;
            ++chars;
          }
          else if (T::lineBreaks() && (CrLf == ind))
          {
            // First must be CR, and it mus be followed by LF
            if ((CR == str[i-1]) && (LF == str[i++]))
            {
              // Ignore the read character
              --lineLength;
              // Verify line length
              if (T::maxLineLength() < lineLength)
                throw std::invalid_argument("Line too long");
              // Should it be exact?
              if (T::fixLineLength() && (T::maxLineLength() > lineLength))
                throw std::invalid_argument("Line too short");
            }
            else
              throw std::invalid_argument("Illegal line break");
            lineLength = 0;
          }
          else if (T::pad() && (Padding == ind))
          {
            // It could be end-of-data padding, in which case this should be 
            // the ultimate or penultimate character
            // If it occurs in the middle, can we ignore it?
            if (!T::ignoreIllegal() && (str.length() > i + 1))
              throw std::invalid_argument("Illegal padding");
          }
        }
        int a;
        // Shift up to fill bucket if data is too short
        for (a = chars; a < 4; ++a)
        {
          bits = (bits << 6);
        }
        // Get bytes out of bucket
        static int shiftlength[] = {16, 8, 0};
        for (a = 0; a < chars - 1; ++a)
        {
          data.push_back((bits >> shiftlength[a]) & 0xff);
        }
      }
    }


  }
}

/*

Copyright (c) 2011, Orjan Westin
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list
of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

The name of the contributor(s) may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

