////////////////////////////////////////////////////////////////////////////////
/*! \file       base64.h
    Header file for functions to encode and decode base64 (RFC 4648)
    According to http://tools.ietf.org/html/rfc4648

    Version:    1.0                       <br>
    Copyright:  Orjan Westin 2011         <br>
    Licence:    BSD (see EOF comment)     <br>
    Contact:    orjan@coolcowstudio.co.uk <br>
    Download:   http://coolcowstudio.co.uk/source/cpp/coding.zip
*///////////////////////////////////////////////////////////////////////////////

#ifndef COCOST_BASE_64_H_
#define COCOST_BASE_64_H_

#include <string>
#include <vector>

//! Cool Cow Studio namespace
namespace CoCoSt 
{
  //! Coding function namespace
  namespace Coding
  {
//! base64 variation specification class
/*! base64 variation specification using template parameters to define behaviour
    and parameters. 

    See http://tools.ietf.org/html/rfc4648

    \tparam Tchar62         Character for index 62 in alphabet
    \tparam Tchar63         Character for index 63 in alphabet             
    \tparam TcharPad        Character to use for padding, or 0 if padding is not used
    \tparaml TfixLineLength false if line length is fixed. If true, maxLineLength is used
    \tparam  TmaxLineLength Maximum (or fixed) line length. 0 if not used. If used, CR+LF is used as line break
    \tparam TignoreIllegal  if false, discards characters not in alphabet; if true, throws error on finding illegal character
*/
    template<
      char Tchar62, 
      char Tchar63,            
      char TcharPad,        
      bool TfixLineLength,  
      int  TmaxLineLength,  
      bool TignoreIllegal>  
    struct base64_variants
    {
      //! Get character for index 62 in alphabet
      static char char62()  
        {return Tchar62;} 
      //! Get character for index 63 in alphabet
      static char char63()  
        {return Tchar63;}          
      //! Get padding character, if any
      static char charPad() 
        {return TcharPad;}
      //! Check if padding is used
      static bool pad()     
        {return TcharPad > 0;}
      //! Check if line length is fixed
      static bool fixLineLength() 
        {return TfixLineLength;}
      //! Get maximum line length
      static int  maxLineLength() 
        {return TmaxLineLength;}
      //! Check if lines should be broken
      static bool lineBreaks()    
        {return TmaxLineLength>0;}
      // Check if illegal characters may be ignored
      static bool ignoreIllegal() 
        {return TignoreIllegal;}
    };
    //! MIME variety of base64 (RFC 2045)
    typedef base64_variants<'+', '/', '=', false, 76, true> base64MIME;

    //! Plain standard (RFC 4648) base64
    typedef base64_variants<'+', '/', '=', false, 0, false> base64;

    //! URL variety of base64 (RFC 4648)
    typedef base64_variants<'-', '_', 0, false, 0, false> base64url;

    //! PEM variety of base64 (RFC 1421)
    typedef base64_variants<'+', '/', '=', true, 64, false> base64PEM;


    template<typename T>
    void bytes_to_base64(const unsigned char* data, 
            size_t length, 
            std::string& str);

    template<typename T>
    void bytes_to_base64(const std::vector<unsigned char>& data, 
            std::string& str);

    template<typename T>
    void base64_to_bytes(const std::string& str, 
            std::vector<unsigned char>& data);

  }
}

#endif // COCOST_BASE_64_H_

/*

Copyright (c) 2011, Orjan Westin
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list
of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

The name of the contributor(s) may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

