/*///////////////////////////////////////////////////////////////////////////////
// \file       hex_str_conv.cpp
    Implementation file for functions converting between hex strings and byte arrays

    Version:    1.0                       <br>
    Copyright:  Orjan Westin 2010         <br>
    Licence:    BSD (see EOF comment)     <br>
    Contact:    orjan@coolcowstudio.co.uk <br>
    Download:   http://coolcowstudio.co.uk/source/cpp/coding.zip
//////////////////////////////////////////////////////////////////////////////*/
#include "hex_str_conv.h"

#include <limits>
#include <cctype>

#include <iostream>

namespace 
{
  // Internal helper function
  inline unsigned char hex_digit_to_nybble(char ch)
  {
    switch (ch)
    {
      case '0': return 0x0;
      case '1': return 0x1;
      case '2': return 0x2;
      case '3': return 0x3;
      case '4': return 0x4;
      case '5': return 0x5;
      case '6': return 0x6;
      case '7': return 0x7;
      case '8': return 0x8;
      case '9': return 0x9;
      case 'a': return 0xa;
      case 'A': return 0xa;
      case 'b': return 0xb;
      case 'B': return 0xb;
      case 'c': return 0xc;
      case 'C': return 0xc;
      case 'd': return 0xd;
      case 'D': return 0xd;
      case 'e': return 0xe;
      case 'E': return 0xe;
      case 'f': return 0xf;
      case 'F': return 0xf;
      default:
      {
        std::string e = "Invalid character in hex string: \'";
        e += ch;
        e += "'";
        throw std::invalid_argument(e);
      }
    }
  }
  // Working with 4-bit nybbles, we can use the value as index to character
  static const std::string::value_type* hexCharactersLC = "0123456789abcdef";
  static const std::string::value_type* hexCharactersUC = "0123456789ABCDEF";
}

namespace CoCoSt
{
  namespace Coding
  {

    // Convert a buffer of bytes to string of corresponding hexadecimal values 
    void bytes_to_hex(const std::vector<unsigned char>& data, std::string& str, HexCase hexCase)
    {
      bytes_to_hex(&data[0], data.size(), str, hexCase);
    }

    // Convert a buffer of bytes to string of corresponding hexadecimal values 
    void bytes_to_hex(const unsigned char* data, size_t length, std::string& str, HexCase hexCase)
    {
      // Sanity check
      static_assert(8 == CHAR_BIT, "bytes_to_hex assumes 8 bits to a char");
      
      // Upper or lower case?
      const std::string::value_type* hexCharacters = (hexCase == lowerCaseHex) ? hexCharactersLC : hexCharactersUC;

      // Clear output
      str.clear();
      
      // No data? Then we're done
      if (0 == length)
  	    return;

      // Result is twice the length of input length
      str.resize(length * 2, ' ');

      for (size_t i = 0; i < length; ++i)
      {
        // High nybble
        str[i<<1] = hexCharacters[(data[i] >> 4) & 0x0f];
        // Low nybble
        str[(i<<1) + 1] = hexCharacters[data[i] & 0x0f];
      }
    }

    // Convert a byte to corresponding hexadecimal value and put in string
    void byte_to_hex(unsigned char data, std::string& str, HexCase hexCase)
    {
      // Sanity check
      static_assert(8 == CHAR_BIT, "bytes_to_hex assumes 8 bits to a char");
      // Upper or lower case?
      const std::string::value_type* hexCharacters = (hexCase == lowerCaseHex) ? hexCharactersLC : hexCharactersUC;
      // High nybble
      str += hexCharacters[(data >> 4) & 0x0f];
      // Low nybble
      str += hexCharacters[data & 0x0f];
    }

    // Convert string of hexadecimal values to corresponding bytes
    void hex_to_bytes(const std::string& str, std::vector<unsigned char>& data)
    {
      // Sanity check
      static_assert(8 == CHAR_BIT, "bytes_to_hex assumes 8 bits to a char");
      
      // Clear output
      data.clear();
      
      // No data? Then we're done
      if (str.empty())
        return;

      // Must be prepared that string can have odd number of 
      // nybbles, in which case the first is treated like the low 
      // nybble of the first byte
      size_t lengthOverflow = str.length() % 2;

      // This also affects the length of the data buffer we
      // allocate (need full  byte for nybble)
      const size_t length = lengthOverflow + str.length() / 2;
      data.resize(length);

      // Buffer for byte conversion
      static char buf[3];
      buf[2] = 0;
      // End of input
      char* pend = &buf[2];

      // Iterators for input and output
      size_t i = 0;
      size_t c = 0;

      // If the first nybble is a low, we'll do it separately
      if (1 == lengthOverflow)
      {
        data[i++] = hex_digit_to_nybble(str[c++]);
      }
      
      // For each output byte, we use two input characters for 
      // high and low nybble, respectively
      while (i < length)
      {
        data[i++] = (hex_digit_to_nybble(str[c++]) << 4) | hex_digit_to_nybble(str[c++]);
      }
    }

    unsigned char hex_to_byte(const std::string& str, unsigned char& data)
    {
      // Sanity check
      static_assert(8 == CHAR_BIT, "bytes_to_hex assumes 8 bits to a char");
      // 
      switch (str.length())
      {
      case 0:
          throw std::invalid_argument("No hex characters given");
          break;
        case 1:
          return hex_digit_to_nybble(str[0]);
          break;
        default:
          return (hex_digit_to_nybble(str[0]) << 4) | hex_digit_to_nybble(str[1]);
          break;
      }
    }

    unsigned char hex_to_byte(std::string::const_iterator& str, const std::string::const_iterator& end)
    {
      // Sanity check
      static_assert(8 == CHAR_BIT, "bytes_to_hex assumes 8 bits to a char");

      if (str == end)
          throw std::invalid_argument("No hex characters given");

      // Assume one nybble first
      unsigned char data = hex_digit_to_nybble(*str++);
      // If there's one more, shift first up
      if (str != end)
        data = (data << 4) | hex_digit_to_nybble(*str++);

      return data;
    }

  }
}

/*

Copyright (c) 2010, Orjan Westin
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list
of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

The name of the contributor(s) may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

