////////////////////////////////////////////////////////////////////////////////
/*! \file       hex_str_conv.h
    Header file for functions converting between hex strings and byte arrays

    Version:    1.0                       <br>
    Copyright:  Orjan Westin 2010         <br>
    Licence:    BSD (see EOF comment)     <br>
    Contact:    orjan@coolcowstudio.co.uk <br>
    Download:   http://coolcowstudio.co.uk/source/cpp/coding.zip
*///////////////////////////////////////////////////////////////////////////////

#ifndef COCOST_HEX_STR_CONV_H_
#define COCOST_HEX_STR_CONV_H_

#include <string>
#include <vector>

//! Cool Cow Studio namespace
namespace CoCoSt 
{
  //! Coding function namespace
  namespace Coding
  {
    //! Indicate whther to use lower or upper case characters in hex output
    enum HexCase
    {
      //! hex characters in lower case (a-f)
      lowerCaseHex,  
      //! hex characters in upper case (A-F) 
      upperCaseHex
    };

    //! Convert a buffer of bytes to string of corresponding hexadecimal values 
    /*! Reads the bytes in the buffer and converts them to corresponding 
	      hexadecimal values. 
		
		    Given a buffer of bytes {7, 233, 57, 42, 198} the string "07e9392ac6" 
		    is generated into the str parameter.
		
        \see bytes_to_hex(const std::vector<unsigned char>& data, std::string& str)
        \param data pointer to memory buffer containing bytes to convert
        \param length number of bytes to read from buffer
        \param str output string into which hexadecimal values will be entered
        \param hexCase indicates whether hex characters should be lower (a-f) or upper (A-F) case
    */
    void bytes_to_hex(const unsigned char* data, 
            size_t length, 
            std::string& str,
            HexCase hexCase = lowerCaseHex);

    //! Convert a buffer of bytes to string of corresponding hexadecimal values 
    /*! Convert a buffer of bytes to string of corresponding hexadecimal values 
        \see bytes_to_hex(const unsigned char* data, size_t length, std::string& str)
        \param data memory buffer containing bytes to convert
        \param str string into which hexadecimal values will be entered
        \param hexCase indicates whether hex characters should be lower (a-f) or upper (A-F) case
    */
    void bytes_to_hex(const std::vector<unsigned char>& data, 
            std::string& str,
            HexCase hexCase = lowerCaseHex);
    
    //! Convert a byte to corresponding hexadecimal value and put in string
    /*! Convert a byte to corresponding hexadecimal value and put in string
        \see bytes_to_hex(const unsigned char* data, size_t length, std::string& str)
        \see bytes_to_hex(const std::vector<unsigned char>& data, std::string& str)
        \param data byte to convert
        \param str string into which hexadecimal value will be entered
        \param hexCase indicates whether hex characters should be lower (a-f) or upper (A-F) case
    */
    void byte_to_hex(unsigned char data, 
            std::string& str,
            HexCase hexCase = lowerCaseHex);

    //! Convert string of hexadecimal values to corresponding bytes
    /*! Convert string of hexadecimal values to corresponding bytes. 
        Takes a string of hex characters ([0-9,a-f,A-F]) and converts every 
        pair of hex characters into a byte, which is put into the data vector. 
        
        Should the string have an odd number of hex characters, the first will 
        be treated as if it had a preceeding '0'.

        Bytes are created and added to data buffer in the order they are read 
        (left-to-right) in the string.
		
		    Throws std::invalid_argument if invalid characters are encountered

        \param str string of hexadecimal values
        \param data output buffer to be filled with bytes
    */
    void hex_to_bytes(const std::string& str, 
            std::vector<unsigned char>& data);

    unsigned char hex_to_byte(const std::string& str);

    unsigned char hex_to_byte(std::string::const_iterator& str, 
            const std::string::const_iterator& end);

  } // namespace Coding
} // namespace CoCoSt

#endif // COCOST_HEX_STR_CONV_H_

/*

Copyright (c) 2010, Orjan Westin
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list
of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

The name of the contributor(s) may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

