/*///////////////////////////////////////////////////////////////////////////////
// \file       quoted_printable.cpp
    Implementation file for functions to encode and decode quoted-printable
    According to http://tools.ietf.org/html/rfc2045#section-6.7

    Version:    1.0                       <br>
    Copyright:  Orjan Westin 2011         <br>
    Licence:    BSD (see EOF comment)     <br>
    Contact:    orjan@coolcowstudio.co.uk <br>
    Download:   http://coolcowstudio.co.uk/source/cpp/coding.zip
//////////////////////////////////////////////////////////////////////////////*/
#include "quoted_printable.h"
#include "hex_str_conv.h"

#include <iostream>

namespace 
{
  // Character constants
  const char CR(0x0d);  // Carriage return
  const char LF(0x0a);  // Line feed
  const char EQ(0x3d);  // Equal (=)
  const char HT(0x09);  // Horizontal tab
  const char SP(0x20);  // Space

  const char TOP(0x7e);   // All above must be encoded

  // Maximum line length (not counting CRLF)
  const int MAXLINELEN(76);

  // Common strings
  const char* CRLF = "\x0d\x0a";
  const char* EQCRLF = "\x3d\x0d\x0a";
}
  
namespace CoCoSt
{
  namespace Coding
  {
    // Lenient in that it accepts lower-case hex
    void qp_to_text(const std::string& qp, std::string& text)
    {
      text.clear();
      text.reserve(qp.length());
      
      int charCount = 0;
      for (std::string::const_iterator i = qp.begin(); i != qp.end();) // Note that increments happen inside loop
      {
        // Special cases first
        switch (*i)
        {
        case EQ:
          {
            // Find what's next
            ++i; 
            ++charCount;

            // Gobble trailing whitespace, if any
            bool trailing = (SP == *i);
            while ((i != qp.end()) && (SP == *i))
            {
              ++i;
              ++charCount;
            }
            // = must be followed by hex or CRLF
            if (i == qp.end())
                throw std::invalid_argument("= may not be the last character"); 

            // Could be soft line break, so check for CRLF 
            if (*i == CR)
            {
              ++i;
              if ((i == qp.end()) || (*i != LF))
                throw std::invalid_argument("CR must be followed by LF"); 
              ++i;
              if (MAXLINELEN < charCount)
                throw std::invalid_argument("Line longer than 76 characters"); 
              charCount = 0;
            }
            else
            {
              // If we've gobbled trailing whitespace, no hex is legal
              if (trailing)
                throw std::invalid_argument("= with trailing space must be followed by CRLF"); 
              // The hex converter accepts single-digit hex values, so one more step of syntax check is needed here
              if (i+1 == qp.end())
                throw std::invalid_argument("= must be followed by two hex digits"); 
              // Convert
              text += static_cast<char>(hex_to_byte(i, qp.end()));
              charCount += 2;
            }
          }
          break;
        case CR:
          // Must be followed by LF
          ++i;
          if ((i == qp.end()) || (*i != LF))
            throw std::invalid_argument("CR must be followed by LF");
          ++i;
          if (MAXLINELEN < charCount)
            throw std::invalid_argument("Line longer than 76 characters"); 
          charCount = 0;
          // Replace with platform-canonical line break
          text += '\n';
          break;
        default:
          // This must be a legal character
          if ((*i < SP) || (*i > TOP))
          {
            // Illegal, unless it's tab and not at eol
            if ((HT != *i) && ((i + 1) != qp.end()) && (*(i + 1) != CR))
              throw std::invalid_argument("Unescaped character");
          }
          text += *i++;
          ++charCount;
        }
      }
    }

    void text_to_qp(const std::string& text, std::string& qp)
    {
      qp.clear();
      qp.reserve(text.length()); // Betting that the size increase is within the safety margin of reserved 
      
      int charCount = 0;
      int nextLineBreak = -2; 
      for (std::string::const_iterator i = text.begin(); i != text.end(); ++i) 
      {
        // Simple case first
        if ((SP < *i) && (TOP >= *i) && (EQ != *i))
        {
          qp += *i;
          ++charCount;
        }
        else
        {
          // Will need to know next line break, if any
          if (-2 == nextLineBreak)
          {
            nextLineBreak = text.find('\n', i - text.begin());
          }
          // If we've hit the line break, put in hard line break
          if (nextLineBreak == (i - text.begin()))
          {
            qp += CRLF;
            charCount = 0;
            nextLineBreak = -2;
          }
          else
          {
            // Space and tab only need encoding before line break
            if (((SP == *i) || (HT == *i)) && (nextLineBreak != (i - text.begin() + 1)))
            {
              qp += *i;
              ++charCount;
            }
            else
            {
              // Encoding needed. First check if it fits
              if ((MAXLINELEN - 3) < charCount)
              {
                // Need to put in soft line break
                qp += EQCRLF;
                charCount = 0;
              }
              std::string hex;
              byte_to_hex(static_cast<unsigned char>(*i), hex, upperCaseHex);
              qp += EQ;
              qp += hex;
              charCount += 3;
            }
          }
        }
        // Check line length
        if ((MAXLINELEN - 1) == charCount)
        {
          // Need to put in soft line break
          qp += EQCRLF;
          charCount = 0;
        }
      }
    }
  }
}

/*

Copyright (c) 2011, Orjan Westin
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list
of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

The name of the contributor(s) may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

